
import pathlib

from .core.app import App


def from_env():
    """
    """
    return API()


class API:
    """
    """

    def get_app(self, name=None, root=None):
        """
        """
        root = pathlib.Path(root) if root else pathlib.Path.cwd()
        name = name or root.name
        app = App(name=name, root=root)
        return app




import io
import os
import pathlib
import logging
import structlog
import subprocess
import tarfile

import docker


logger = structlog.get_logger(__name__)


class Environment:
    """
    Environment

    This class represents a container-based environment.
    """

    def __init__(self, image):
        self._image = image
        self._container = None
        self._volumes = {}
        self.workdir = "/home/cuta"
        self._docker = None

    @property
    def docker(self):
        if not self._docker:
            self._docker = docker.from_env()

        return self._docker

    @property
    def image(self):
        return self._image

    @property
    def container(self):
        return self._container

    @property
    def volumes(self):
        return self._volumes

    @property
    def status(self):
        return self.container.status if self.container else None

    @property
    def active(self):
        """
        Check if the environment is active.

        The environment is active when the container for it
        is running.
        """
        return self.status == "running"

    def mount(self, source, target=None, mode="rw"):
        """
        Mount source directory
        """
        target = target or "/home/cuta"
        self._volumes[source] = {"bind": target, "mode": mode}

    def unmount(self, source):
        """
        Unmount source directory
        """
        self._volumes.pop(source, None)

    def activate(self):
        """
        Activate

        Activate the environment (i.e. start container)
        """
        if self.active:
            return

        command = ["sleep", "infinity"]
        self._container = self.docker.containers.create(
            self.image,
            command=command,
            volumes=self.volumes,
            working_dir=self.workdir,
            auto_remove=True,
            )
        self._container.start()
        self._container.reload()

        logger.info(f"Environment activated", container_id=self.container.id)

    def deactivate(self):
        """
        Deactive

        Deactivate the shell (i.e. kill and remove container)
        """
        if not self.active:
            return

        container_id = self.container.id
        self.container.kill()
        self._container = None

        logger.info(f"Environment deactivated", container_id=container_id)

    def __enter__(self):
        self.activate()
        return self

    def __exit__(self, *exc_details):
        self.deactivate()

    def exec(self, cmd, **kwargs):
        """
        Execute the given command

        Execute the given command inside the container.
        """
        if not self.container:
            raise RuntimeError("Environment must be activated")

        return self.container.exec_run(cmd, **kwargs)

    def get_file(self, source, target):
        """
        Get file

        Copy source file to the given target location.
        """
        if not self.container:
            raise RuntimeError("Environment must be activated")

        # Get archive (tar) from the container.
        stream = io.BytesIO()
        chunks, stat = self.container.get_archive(str(source))
        for chunk in chunks:
            stream.write(chunk)

        # Move cursor back to the beginning of the file.
        stream.seek(0)

        # Extract file data and write it to the target.
        source = pathlib.Path(source)
        target = pathlib.Path(target)
        tar = tarfile.TarFile(fileobj=stream)
        extracted = tar.extractfile(source.name)
        with target.open("wb") as file_:
            data = extracted.read()
            file_.write(data)

        logger.info(f"Copied {self.container.id}:{source} to {target}")

        return target


def pull(name):
    """
    Pull environment from remote repository (i.e. DockerHub)
    """
    image = get_image(name)
    if not image:
        client = docker.from_env()
        output = client.api.pull(name, stream=True, decode=True)
        for item in output:
            yield item


def exists(name):
    """
    Check if the given environment exists locally
    """
    return bool(get_env(name))


def get_env(name):
    """
    Get environment with the given name/id.
    """
    image = get_image(name)
    return Environment(image) if image else None


def get_image(name):
    """
    Get docker image with the given name/id.
    """
    client = docker.from_env()
    try:
        image = client.images.get(name)
    except docker.errors.ImageNotFound as err:
        image = None

    return image




from emoji import emojize
import sys
import typer

import cuta


_color = None


def get_color():
    return _color


def set_color(color):
    global _color
    _color = color


def echo(*args, **kwargs):
    text = fmt(*args, **kwargs)
    typer.echo(text)


def fmt(msg, color=None, emoji=None, **options):
    color = color or get_color()
    if color or options:
        msg = typer.style(msg, fg=color, **options)

    if emoji:
        msg = emojize(f":{emoji}: {msg}", use_aliases=True)

    return msg


def version():
    msg = f"Cuta: {cuta.version}"
    echo(msg, color=typer.colors.BRIGHT_WHITE, bold=True)


def title(title, color=None, emoji=None):
    echo("-" * 72, color=color)
    echo(title, color=color, emoji=emoji)
    echo("-" * 72, color=color)


def complete(msg="Complete!"):
    echo(msg, color=typer.colors.BRIGHT_WHITE, bold=True)


def clear_lines(count=1):
    for index in range(count):
        sys.stdout.write("\x1b[1A\x1b[2K")




import toml
from typing import Any

"""
Toml File

This module contains a utility class for working with TOML files.
"""


class TomlFile:
    """
    TOML File

    This class represents a TOML file and provides
    some useful properties/methods for accessing the
    file data.
    """

    def __init__(self, path: str, data: dict = None):
        self._path = path
        self._data = data or {}

    @property
    def path(self) -> str:
        return self._path

    @property
    def data(self) -> dict:
        return self._data

    def get_value(self, path: str) -> Any:
        """
        Get value

        Get an attribute of the toml file. Nested
        values can be accessed using a dotted key.
        """
        data = self.data
        keys = path.split(".")
        for key in keys:
            if isinstance(data, dict):
                data = data[key]

        return data

    def set_value(self, path: str, value: Any) -> None:
        """
        Set value

        Set an attribute of the toml file. Nested
        values can be accessed using a dotted key.
        """
        data = self.data
        keys = path.split(".")
        parent = None
        for key in keys:
            if isinstance(data, dict):
                parent = data
                data = data[key]

        parent[key] = value

    def read(self) -> None:
        """
        Read data from disk.
        """
        self._data = read_toml(str(self.path))

    def write(self) -> None:
        """
        Write data to disk.
        """
        write_toml(str(self.path), self.data)


def read_toml(path: str) -> dict:
    """
    Read project file (i.e. pyproject.toml)
    """
    with open(path, "r") as project_file:
        return toml.load(project_file)


def write_toml(path: str, data: dict) -> None:
    """
    Write project file (i.e. pyproject.toml)
    """
    with open(path, "w") as project_file:
        toml.dump(data, project_file)




import emoji
import re
import sys
import typer
from typing import List

import cuta

from . import utils


app = typer.Typer()


@app.command("init")
def init():
    """
    Initialize application.
    """
    utils.version()

    api = cuta.from_env()
    app = api.get_app()

    create_env(app)
    create_package(app)

    utils.complete()


def create_env(app):
    """
    Create base environment
    """
    tag = f"ylathouris/cuta:{app.platform}-{app.runtime}"
    if cuta.core.env.exists(tag):
        return

    color = utils.get_color()
    utils.set_color(typer.colors.BRIGHT_CYAN)
    utils.title("Creating Base Environment", emoji="whale")

    report = {}
    for item in cuta.core.env.pull(app.base):
        uid = item.get("id", "")
        status = item.get("status", "")
        progress = item.get("progress")
        lines = []
        if progress:
            report[uid] = f"{uid}: {status} {progress}"
            lines = report.values()
            count = len(lines)
            for line in lines:
                utils.echo(line)

            utils.clear_lines(count)

    utils.set_color(color)


def create_package(app):
    """
    Initialize package config/files.
    """
    color = utils.get_color()
    utils.set_color(typer.colors.BRIGHT_YELLOW)
    utils.title("Creating Package", emoji="package")

    for output in app.init():
        utils.echo(output)

    utils.set_color(color)


@app.command("add")
def add_deps(deps: List[str], dev: bool = typer.Option(False, "--dev")):
    echo_version()
    api = cuta.from_env()
    app = api.get_app()

    typer.echo(yellow(emoji.emojize(":package: Adding dependencies...")))
    for output in app.add(deps, dev=dev):
        typer.echo(yellow("Packaging: ") + output)

    done()


@app.command("remove")
def remove_deps(deps: List[str], dev: bool = typer.Option(False, "--dev")):
    echo_version()
    api = cuta.from_env()
    app = api.get_app()

    typer.echo(yellow(emoji.emojize(":package: Removing dependencies...")))
    for output in app.remove(deps, dev=dev):
        typer.echo(yellow("Packaging: ") + output)

    done()


@app.command("build")
def build():
    echo_version()
    api = cuta.from_env()
    app = api.get_app()

    typer.echo(green(emoji.emojize(f":package: Building app: {app.tag}")))
    for output in app.build():
        typer.echo(cyan("Docker: ") + output)

    done()


@app.command("shell")
def shell():
    echo_version()
    typer.echo(green(emoji.emojize(f":whale: Initializing cuta shell")))
    api = cuta.from_env()
    app = api.get_app()
    app.shell()


@app.command("delete")
def delete():
    echo_version()
    api = cuta.from_env()
    app = api.get_app()
    typer.echo(yellow(emoji.emojize(f":package: Deleting app: {app.name}")))
    app.delete()
    done()


def echo_version():
    """
    """
    msg = green(f"Cuta Version: {cuta.version}")
    typer.echo(msg)


def cyan(msg: str) -> str:
    return typer.style(msg, fg=typer.colors.BRIGHT_CYAN)


def blue(msg: str) -> str:
    return typer.style(msg, fg=typer.colors.BRIGHT_BLUE)


def green(msg: str) -> str:
    return typer.style(msg, fg=typer.colors.BRIGHT_GREEN)


def yellow(msg: str) -> str:
    return typer.style(msg, fg=typer.colors.BRIGHT_YELLOW)


def done(msg="Complete!"):
    text = typer.style(msg, fg=typer.colors.BRIGHT_WHITE, bold=True)
    typer.echo(text)




import codecs
import pathlib
import re
import shutil
import structlog
import subprocess

import docker

from .env import get_env


logger = structlog.get_logger(__name__)


class App:
    """
    Application

    This class represents an application.
    """

    def __init__(self, name=None, root=None):
        self._name = name
        self._root = pathlib.Path(root) if root else None
        self.platform = "aws"
        self.runtime = "python3.8"
        self.version = "0.1.0"
        self._envs = {}

    @property
    def name(self):
        return self._name

    @property
    def root(self):
        return self._root

    @property
    def config(self):
        return self.root.joinpath(".cuta")

    @property
    def tag(self):
        return f"{self.name}:{self.version}"

    @property
    def base(self):
        return f"ylathouris/cuta:{self.platform}-{self.runtime}"

    def init(self):
        """
        Initialize

        Initialize new application using the current working
        directory. This method will create the required package
        files for the app.
        """
        self._root = self._root or pathlib.Path.cwd()
        self._name = self._name or self.root.name

        logger.info(f"App Context", platform=self.platform, runtime=self.runtime)
        for output in self.create_package():
            yield output

        self.create_dockerfile()

    def create_package(self):
        """
        Create package
        """
        path = self.config.joinpath("pyproject.toml")
        if path.exists():
            return

        with self.env() as env:
            logger.debug(f"Creating package", root=str(self.root), name=self.name)
            cmd = ["poetry", "init", f"--no-interaction", f"--name={self.name}"]
            exit_code, output = env.exec(cmd, stream=True)
            for chunk in output:
                yield chunk.decode("utf-8").strip()

            logger.debug(f"Locking package")
            env.exec(["poetry", "lock"])

    def create_dockerfile(self):
        """
        Create dockerfile
        """
        root = pathlib.Path(__file__).parent.parent.parent.parent
        template = root.joinpath("template", "Dockerfile")
        dockerfile = self.config.joinpath("Dockerfile")
        logger.debug(f"Copying Docker template")
        shutil.copyfile(template, dockerfile)

    def add(self, deps, dev=False):
        """
        Add dependencies

        Add dependencies required by the app. If `dev=True`, the
        dependencies will be added as development only dependencies
        and will not be present in the final build.
        """
        logger.info("Adding dependencies", dev=dev, deps=deps)
        with self.env() as env:
            cmd = ["poetry", "add", *deps]
            if dev:
                cmd.insert(2, "--dev")

            exit_code, output = env.exec(cmd, stream=True)
            for chunk in output:
                chunk = chunk.decode("utf-8")
                for line in chunk.splitlines():
                    yield line

    def remove(self, deps, dev=False):
        """
        Remove dependencies

        Remove dependencies required by the app.
        """
        logger.info("Removing dependencies", dev=dev, deps=deps)
        with self.env() as env:
            cmd = ["poetry", "remove", *deps]
            if dev:
                cmd.insert(2, "--dev")

            exit_code, output = env.exec(cmd, stream=True)
            for chunk in output:
                yield chunk.decode("utf-8").strip()

    def env(self):
        tag = f"ylathouris/cuta:{self.platform}-{self.runtime}"
        env = get_env(tag)
        env.workdir = f"/home/cuta/{self.name}/.cuta/"
        env.mount(self.root, f"/home/cuta/{self.name}")
        return env

    def build(self):
        """
        """
        dockerdir = self.root.joinpath(".cuta")
        client = docker.from_env()
        for output in client.api.build(tag=self.tag, path=str(dockerdir)):
            output = output.decode("utf-8")
            output = output.strip()
            output = codecs.decode(output, "unicode_escape")
            match = re.search(":\"(.*)\S?", output, re.M)
            if match:
                yield match.group(1)
            else:
                yield repr(output)

    def shell(self):
        """
        """
        subprocess.run([
            "docker",
            "run",
            "--interactive",
            "--tty",
            "--rm",
            self.tag,
            "bash"
        ])

    def delete(self):
        """
        """
        if self.config.exists():
            logger.info(f"Deleting cuta config", path=str(self.config))
            shutil.rmtree(str(self.config))

        try:
            logger.debug(f"Removing build", tag=self.tag)
            client = docker.from_env()
            client.images.remove(self.tag, force=True)
        except docker.errors.ImageNotFound as err:
            pass


class PackageManager:
    """
    """

    def __init__(self, session):
        pass


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'source'}

packages = \
['cuta', 'cuta.cli', 'cuta.core']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.3,<0.5.0',
 'docker>=4.2.0,<5.0.0',
 'emoji>=0.5.4,<0.6.0',
 'structlog>=20.1.0,<21.0.0',
 'typer>=0.2.1,<0.3.0']

entry_points = \
{'console_scripts': ['cuta = cuta.cli:base']}

setup_kwargs = {
    'name': 'cuta',
    'version': '0.1.1',
    'description': '',
    'long_description': '[![CircleCI](https://circleci.com/gh/ylathouris/bio.svg?style=shield)](https://circleci.com/gh/ylathouris/rolodex)  ![Coverage](coverage.svg)\n\n---\n\n# Cuta\n\nCuta is a set of tools designed to make building applications easier.\n\n\n<br/>\n\n## Next Steps\n\n* [Installation](docs/installation.md)\n* [Quick Start](docs/quickstart.md)\n\n<br/>\n\n\n[Docker]: https://www.docker.com/\n[Terraform]: https://www.terraform.io/',
    'author': 'Yani Lathouris',
    'author_email': 'ylathouris@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ylathouris/cuta',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
